import pathlib
from typing import Any, List, Optional, Tuple
from unittest import mock

import pytest
from click.testing import CliRunner
from pyfakefs.fake_filesystem import FakeFilesystem

from tests.decorators import test_connector

from gcs_migration.commands.create import create
from gcs_migration.models import MigrationPlan
from gcs_migration.parsers import (
    BlackPearlMapEntry,
    CephConfig,
    GCSConfig,
    GridFTPOptionParser,
    GsiAuthzParser,
    HPSSConfig,
    S3Config,
    S3MapfileParser,
)
from gcs_migration.util import AuthorizeSharing


@pytest.mark.no_connector_only
def test_create(
    fs: FakeFilesystem,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    host_endpoint_uuid: str,
    blackpearl_endpoint: str,
    blackpearl_access_id_filename: str,
    blackpearl_map_entries: List[BlackPearlMapEntry],
    hpss_config: HPSSConfig,
    mock_s3_config: S3Config,
    mock_s3_mapfile: S3MapfileParser,
    mock_ceph_config: CephConfig,
) -> None:
    tmp_path = pathlib.Path(".")
    ret_migration_plan = MigrationPlan(tmp_path)

    with mock.patch(
        "gcs_migration.commands.create.GCSConfig",
        return_value=gcs_config,
    ), mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser",
        return_value=gridftp_config_parser,
    ), mock.patch(
        "gcs_migration.commands.create.MigrationPlan",
        return_value=ret_migration_plan,
    ), mock.patch.object(
        AuthorizeSharing,
        "map_and_authorize_sharing",
        return_value=authorization_response_success,
    ), mock.patch(
        "click.prompt", return_value=host_endpoint_uuid
    ), mock.patch(
        "gcs_migration.commands.create.BlackPearlConfig"
    ) as mock_blackpearl_config, mock.patch(
        "gcs_migration.models.storage_gateway.BlackPearlAccessIdFileParser"
    ) as mock_bp_access_id_parser, mock.patch(
        "gcs_migration.commands.create.S3Config", return_value=mock_s3_config
    ), mock.patch(
        "gcs_migration.models.storage_gateway.S3MapfileParser",
        return_value=mock_s3_mapfile,
    ), mock.patch(
        "gcs_migration.commands.create.CephConfig", return_value=mock_ceph_config
    ), mock.patch(
        "gcs_migration.commands.create.HPSSConfig"
    ) as mock_hpss_config:
        mock_blackpearl_config.return_value.endpoint = blackpearl_endpoint
        mock_blackpearl_config.return_value.access_id_file = (
            blackpearl_access_id_filename
        )
        mock_access_id_parser = mock_bp_access_id_parser.return_value
        mock_access_id_parser.map_entries = {
            k.username: k for k in blackpearl_map_entries
        }
        mock_bp_access_id_parser.return_value.__contains__.side_effect = (
            lambda x: x in mock_access_id_parser.map_entries
        )
        mock_bp_access_id_parser.return_value.__getitem__.side_effect = (
            lambda x: mock_access_id_parser.map_entries[x]
        )
        mock_hpss_config.return_value = hpss_config

        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 0
        loaded_plan = MigrationPlan.load(tmp_path)

        assert (
            loaded_plan.gsi_authz_configuration
            == migration_plan_cmd_create.gsi_authz_configuration
        )
        assert loaded_plan.storage_gateway == migration_plan_cmd_create.storage_gateway
        assert (
            loaded_plan.mapped_collection == migration_plan_cmd_create.mapped_collection
        )
        assert (
            loaded_plan.guest_collections == migration_plan_cmd_create.guest_collections
        )

        assert loaded_plan == ret_migration_plan


@test_connector("file")
def test_create_no_config(
    fs: FakeFilesystem,
    connector: str,
) -> None:
    runner = CliRunner()
    result = runner.invoke(create)
    assert result.exit_code == 1
    assert "This command must be run on a GCSv4 node" in result.output


@test_connector("file")
def test_create_unreadable_gridftp_d(
    fs: FakeFilesystem,
    connector: str,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    host_endpoint_uuid: str,
) -> None:
    with mock.patch(
        "gcs_migration.commands.create.GCSConfig",
        return_value=gcs_config,
    ), mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser",
        side_effect=PermissionError("Permission denied: '/etc/gridftp.d'"),
    ):
        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 1
        assert "Unable to parse gridftp" in result.output


@test_connector("file")
def test_create_unreadable_gsi_authz_conf(
    fs: FakeFilesystem,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    host_endpoint_uuid: str,
) -> None:
    tmp_path = pathlib.Path(".")

    with mock.patch(
        "gcs_migration.commands.create.GCSConfig",
        return_value=gcs_config,
    ), mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser",
        return_value=gridftp_config_parser,
    ), mock.patch(
        "gcs_migration.commands.create.MigrationPlan"
    ) as mock_migration_plan, mock.patch.object(
        GsiAuthzParser,
        "parse",
        side_effect=PermissionError(
            "Permission denied: '/etc/gridmap_verify_myproxy_callout-gsi_authz.conf'"
        ),
    ), mock.patch(
        "click.prompt", return_value=host_endpoint_uuid
    ):
        ret_migration_plan = MigrationPlan(tmp_path)
        mock_migration_plan.return_value = ret_migration_plan

        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 1
        assert "Permission denied" in result.output


@test_connector("blackpearl")
def test_create_unreadable_blackpearl_config(
    fs: FakeFilesystem,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    host_endpoint_uuid: str,
) -> None:
    tmp_path = pathlib.Path(".")

    with mock.patch(
        "gcs_migration.commands.create.GCSConfig"
    ) as mock_gcs_config, mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser"
    ) as mock_gridftp_option_parser, mock.patch(
        "gcs_migration.commands.create.MigrationPlan"
    ) as mock_migration_plan, mock.patch.object(
        AuthorizeSharing, "map_and_authorize_sharing"
    ) as mock_map_and_authorize_sharing, mock.patch(
        "click.prompt", return_value=host_endpoint_uuid
    ), mock.patch(
        "gcs_migration.commands.create.BlackPearlConfig",
        side_effect=PermissionError(
            "Permission denied: '/etc/blackpearl/GridFTPConfig'"
        ),
    ):
        mock_gcs_config.return_value = gcs_config
        mock_gridftp_option_parser.return_value = gridftp_config_parser
        mock_map_and_authorize_sharing.return_value = authorization_response_success
        ret_migration_plan = MigrationPlan(tmp_path)
        mock_migration_plan.return_value = ret_migration_plan

        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 1
        assert "/etc/blackpearl/GridFTPConfig" in result.output


@test_connector("hpss_local")
def test_create_unreadable_hpss_local_config(
    fs: FakeFilesystem,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    host_endpoint_uuid: str,
) -> None:
    tmp_path = pathlib.Path(".")

    with mock.patch(
        "gcs_migration.commands.create.GCSConfig"
    ) as mock_gcs_config, mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser"
    ) as mock_gridftp_option_parser, mock.patch(
        "gcs_migration.commands.create.MigrationPlan"
    ) as mock_migration_plan, mock.patch.object(
        AuthorizeSharing, "map_and_authorize_sharing"
    ) as mock_map_and_authorize_sharing, mock.patch(
        "click.prompt", return_value=host_endpoint_uuid
    ), mock.patch(
        "gcs_migration.commands.create.HPSSConfig",
        side_effect=PermissionError(
            "Permission denied: '/var/hpss/etc/gridftp_hpss_dsi.conf'"
        ),
    ):
        mock_gcs_config.return_value = gcs_config
        mock_gridftp_option_parser.return_value = gridftp_config_parser
        mock_map_and_authorize_sharing.return_value = authorization_response_success
        ret_migration_plan = MigrationPlan(tmp_path)
        mock_migration_plan.return_value = ret_migration_plan

        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 1
        assert "/var/hpss/etc/gridftp_hpss_dsi.conf" in result.output


@test_connector("file")
def test_create_no_mapping_for_user(
    fs: FakeFilesystem,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_failure: Tuple[Optional[str], str, int],
    host_endpoint_uuid: str,
) -> None:
    tmp_path = pathlib.Path(".")
    ret_migration_plan = MigrationPlan(tmp_path)

    with mock.patch(
        "gcs_migration.commands.create.GCSConfig",
        return_value=gcs_config,
    ), mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser",
        return_value=gridftp_config_parser,
    ), mock.patch(
        "gcs_migration.commands.create.MigrationPlan",
        return_value=ret_migration_plan,
    ), mock.patch.object(
        AuthorizeSharing,
        "map_and_authorize_sharing",
        return_value=authorization_response_failure,
    ), mock.patch(
        "click.prompt", return_value=host_endpoint_uuid
    ):
        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 1
        assert "Unable to determine local user for" in result.output


@test_connector("s3")
def test_create_s3_no_entry(
    fs: FakeFilesystem,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    host_endpoint_uuid: str,
    mock_s3_config: S3Config,
) -> None:
    tmp_path = pathlib.Path(".")
    ret_migration_plan = MigrationPlan(tmp_path)

    with mock.patch(
        "gcs_migration.commands.create.GCSConfig",
        return_value=gcs_config,
    ), mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser",
        return_value=gridftp_config_parser,
    ), mock.patch(
        "gcs_migration.commands.create.MigrationPlan",
        return_value=ret_migration_plan,
    ), mock.patch.object(
        AuthorizeSharing,
        "map_and_authorize_sharing",
        return_value=authorization_response_success,
    ), mock.patch(
        "click.prompt", return_value=host_endpoint_uuid
    ), mock.patch(
        "gcs_migration.commands.create.S3Config", return_value=mock_s3_config
    ), mock.patch(
        "gcs_migration.models.storage_gateway.S3MapfileParser",
        return_value={authorization_response_success[0]: None},
    ):
        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 0
        assert "Unable to determine s3 key" in result.output


@test_connector("blackpearl")
def test_create_blackpearl_no_entry(
    fs: FakeFilesystem,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    host_endpoint_uuid: str,
    blackpearl_endpoint: str,
    blackpearl_access_id_filename: str,
    blackpearl_map_entries: List[BlackPearlMapEntry],
) -> None:
    tmp_path = pathlib.Path(".")
    ret_migration_plan = MigrationPlan(tmp_path)

    with mock.patch(
        "gcs_migration.commands.create.GCSConfig",
        return_value=gcs_config,
    ), mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser",
        return_value=gridftp_config_parser,
    ), mock.patch(
        "gcs_migration.commands.create.MigrationPlan",
        return_value=ret_migration_plan,
    ), mock.patch.object(
        AuthorizeSharing,
        "map_and_authorize_sharing",
        return_value=authorization_response_success,
    ), mock.patch(
        "click.prompt", return_value=host_endpoint_uuid
    ), mock.patch(
        "gcs_migration.commands.create.BlackPearlConfig"
    ) as mock_blackpearl_config, mock.patch(
        "gcs_migration.models.storage_gateway.BlackPearlAccessIdFileParser"
    ) as mock_bp_access_id_parser:
        mock_blackpearl_config.return_value.endpoint = blackpearl_endpoint
        mock_blackpearl_config.return_value.access_id_file = (
            blackpearl_access_id_filename
        )
        mock_access_id_parser = mock_bp_access_id_parser.return_value
        mock_access_id_parser.map_entries = {authorization_response_success[0]: None}
        mock_bp_access_id_parser.return_value.__contains__.side_effect = (
            lambda x: x in mock_access_id_parser.map_entries
        )
        mock_bp_access_id_parser.return_value.__getitem__.side_effect = (
            lambda x: mock_access_id_parser.map_entries[x]
        )

        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 0
        assert "Unable to determine BlackPearl" in result.output


@test_connector("blackpearl")
def test_create_blackpearl_unreadable_access_id_file(
    fs: FakeFilesystem,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    host_endpoint_uuid: str,
    blackpearl_endpoint: str,
    blackpearl_access_id_filename: str,
    blackpearl_map_entries: List[BlackPearlMapEntry],
) -> None:
    tmp_path = pathlib.Path(".")
    ret_migration_plan = MigrationPlan(tmp_path)

    with mock.patch(
        "gcs_migration.commands.create.GCSConfig",
        return_value=gcs_config,
    ), mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser",
        return_value=gridftp_config_parser,
    ), mock.patch(
        "gcs_migration.commands.create.MigrationPlan",
        return_value=ret_migration_plan,
    ), mock.patch.object(
        AuthorizeSharing,
        "map_and_authorize_sharing",
        return_value=authorization_response_success,
    ), mock.patch(
        "click.prompt", return_value=host_endpoint_uuid
    ), mock.patch(
        "gcs_migration.commands.create.BlackPearlConfig"
    ) as mock_blackpearl_config, mock.patch(
        "gcs_migration.models.storage_gateway.BlackPearlAccessIdFileParser",
        side_effect=PermissionError("Permission denied: '/etc/blackpearl/AccessId'"),
    ):
        mock_blackpearl_config.return_value.endpoint = blackpearl_endpoint
        mock_blackpearl_config.return_value.access_id_file = (
            blackpearl_access_id_filename
        )

        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 1
        assert "Unable to read BlackPearl access" in result.output


def test_create_v5_endpoint(
    fs: FakeFilesystem,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    v5_endpoint_uuid: str,
    host_endpoint_uuid: str,
) -> None:
    tmp_path = pathlib.Path(".")

    with mock.patch(
        "gcs_migration.commands.create.GCSConfig"
    ) as mock_gcs_config, mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser"
    ) as mock_gridftp_option_parser, mock.patch(
        "gcs_migration.commands.create.MigrationPlan"
    ) as mock_migration_plan, mock.patch.object(
        AuthorizeSharing, "map_and_authorize_sharing"
    ) as mock_map_and_authorize_sharing, mock.patch(
        "click.prompt", return_value=host_endpoint_uuid
    ):
        mock_gcs_config.return_value = gcs_config
        gcs_config.endpoint_id = v5_endpoint_uuid
        mock_gridftp_option_parser.return_value = gridftp_config_parser
        mock_map_and_authorize_sharing.return_value = authorization_response_success
        ret_migration_plan = MigrationPlan(tmp_path)
        mock_migration_plan.return_value = ret_migration_plan

        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 1
        assert "is a GCS v5 endpoint" in result.output


def test_create_v5_node(
    fs: FakeFilesystem,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    host_endpoint_uuid: str,
) -> None:
    tmp_path = pathlib.Path(".")
    with mock.patch(
        "gcs_migration.commands.create.GCSConfig"
    ) as mock_gcs_config, mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser"
    ) as mock_gridftp_option_parser, mock.patch(
        "gcs_migration.commands.create.MigrationPlan"
    ) as mock_migration_plan, mock.patch.object(
        AuthorizeSharing, "map_and_authorize_sharing"
    ) as mock_map_and_authorize_sharing, mock.patch(
        "click.prompt", return_value=host_endpoint_uuid
    ):
        mock_gcs_config.return_value = gcs_config
        mock_gridftp_option_parser.return_value = gridftp_config_parser
        gridftp_config_parser.options["version_tag"] = "GCS-5.4.50"
        mock_map_and_authorize_sharing.return_value = authorization_response_success
        ret_migration_plan = MigrationPlan(tmp_path)
        mock_migration_plan.return_value = ret_migration_plan

        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 1
        assert "GridFTP server configuration is for a GCSv5 endpoint" in result.output


def test_create_v5_collection(
    fs: FakeFilesystem,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    v5_collection_uuid: str,
    host_endpoint_uuid: str,
) -> None:
    tmp_path = pathlib.Path(".")
    with mock.patch(
        "gcs_migration.commands.create.GCSConfig"
    ) as mock_gcs_config, mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser"
    ) as mock_gridftp_option_parser, mock.patch(
        "gcs_migration.commands.create.MigrationPlan"
    ) as mock_migration_plan, mock.patch.object(
        AuthorizeSharing, "map_and_authorize_sharing"
    ) as mock_map_and_authorize_sharing, mock.patch(
        "click.prompt", return_value=host_endpoint_uuid
    ):
        mock_gcs_config.return_value = gcs_config
        gcs_config.endpoint_id = v5_collection_uuid
        mock_gridftp_option_parser.return_value = gridftp_config_parser
        mock_map_and_authorize_sharing.return_value = authorization_response_success
        ret_migration_plan = MigrationPlan(tmp_path)
        mock_migration_plan.return_value = ret_migration_plan

        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 1
        assert "is a GCS v5 collection" in result.output


@test_connector("file")
def test_create_cant_write_plan(
    fs: FakeFilesystem,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    host_endpoint_uuid: str,
) -> None:
    tmp_path = pathlib.Path(".")
    ret_migration_plan = MigrationPlan(tmp_path)

    with mock.patch(
        "gcs_migration.commands.create.GCSConfig",
        return_value=gcs_config,
    ), mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser",
        return_value=gridftp_config_parser,
    ), mock.patch(
        "gcs_migration.commands.create.MigrationPlan",
        return_value=ret_migration_plan,
    ), mock.patch.object(
        AuthorizeSharing,
        "map_and_authorize_sharing",
        return_value=authorization_response_success,
    ), mock.patch(
        "click.prompt", return_value=host_endpoint_uuid
    ), mock.patch.object(
        MigrationPlan,
        "dump",
        side_effect=PermissionError("Permission denied: 'migration_plan/version.yaml'"),
    ):
        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 1
        assert "Unable to write migration plan" in result.output


def test_requires_logged_in(mock_not_logged_in: None) -> None:
    runner = CliRunner()
    result = runner.invoke(create, catch_exceptions=False)
    assert "you will have to login" in result.output


@pytest.mark.parametrize("connector", ["ceph"])
def test_create_ceph_config_unreadable(
    fs: FakeFilesystem,
    mock_get_transfer_api: Any,
    gcs_config: GCSConfig,
    gridftp_config_parser: GridFTPOptionParser,
    migration_plan_cmd_create: MigrationPlan,
    authorization_response_success: Tuple[Optional[str], str, int],
    host_endpoint_uuid: str,
) -> None:
    tmp_path = pathlib.Path(".")

    with mock.patch(
        "gcs_migration.commands.create.GCSConfig"
    ) as mock_gcs_config, mock.patch(
        "gcs_migration.commands.create.GridFTPOptionParser"
    ) as mock_gridftp_option_parser, mock.patch(
        "gcs_migration.commands.create.MigrationPlan"
    ) as mock_migration_plan, mock.patch.object(
        AuthorizeSharing, "map_and_authorize_sharing"
    ) as mock_map_and_authorize_sharing, mock.patch(
        "click.prompt", return_value=host_endpoint_uuid
    ), mock.patch(
        "gcs_migration.commands.create.CephConfig",
        side_effect=PermissionError(
            "Permission denied: '/etc/globus/globus-gridftp-server-ceph.conf'"
        ),
    ):
        mock_gcs_config.return_value = gcs_config
        mock_gridftp_option_parser.return_value = gridftp_config_parser
        mock_map_and_authorize_sharing.return_value = authorization_response_success
        ret_migration_plan = MigrationPlan(tmp_path)
        mock_migration_plan.return_value = ret_migration_plan

        runner = CliRunner()
        result = runner.invoke(create)
        assert result.exit_code == 1
        assert "Permission denied" in result.output
